# ----------------------
# Space Station Lockout
# In this problem you are to build a planner that finds a solution for a
# Space Station Lockout game.
#
# In this problem, a game is represented by an array indicating the position
# of the carbon based unit and its bot helpers:
#
#  [[1,1], [0, 1], [0, 2], [3, 3], [4,1]]
#
#  The first element of the array indicates the carbon-based unit and the other
#  elements are the bot helpers. This array represents the following position
#  where X represents the carbon-based unit, 1-4 are the bot helpers and the O
#  indicates the airlock:
#
#   +---+---+---+---+---+
#   |   | 1 | 2 |   |   |
#   +---+---+---+---+---+
#   |   | X |   |   |   |
#   +---+---+---+---+---+
#   |   |   | O |   |   |
#   +---+---+---+---+---+
#   |   |   |   | 3 |   |
#   +---+---+---+---+---+
#   |   | 4 |   |   |   |
#   +---+---+---+---+---+
#
#  The playing board is 5 x 5 squares. The idea is that you on the outside of
#  a  space station with your bots; an emergency occurred and you must reach
#  the escape hatch (the center square) to return to the safety of the station.
#  A piece can move up, down, left, or right. A piece can only stop by being
#  blocked by another piece. If it is not blocked it flies off the edge of the
#  board. If the piece is stopped at the airlock, it is removed from the board.
#  Again, the goal of the game is for the carbon unit to reach the airlock.
#
#  A solution to the problem is given as a list of moves. For example, a solution
#  to the above problem is
#
#   [[0, 'South'], [0, 'East'], [0, 'North'], [1, 'South'], [3, 'West'], [0, 'South']]
#
#  meaning the carbon-based unit goes South and then East, and North.Then Unit 1 moves
#  South, and so on.
#  
#  -----------------------------------
#
#  Programmer Instructions:
#
#  This is to be done individually without outside help.
#
#  Write the method getSolution which starts on line 170. The function should return a
#  list of moves.  (180XP)
#
#  A PriorityQueue class is provided for you
#  The getSuccessors method returns a list of successor states to a particular
#  game state
#
#  Optional (worth 70XP): write a method checkSolution that takes a game and a solution
#  as arguments and returns True if the solution is indeed a solution to the game
#  and False otherwise

import heapq
class PriorityQueue:
  """
    Implements a priority queue data structure. Each inserted item
    has a priority associated with it and the client is usually interested
    in quick retrieval of the lowest-priority item in the queue. This
    data structure allows O(1) access to the lowest-priority item.
    
    Note that this PriorityQueue does not allow you to change the priority
    of an item.  However, you may insert the same item multiple times with
    different priorities.
  """  
  def  __init__(self):  
    self.heap = []
    
  def push(self, item, priority):
      pair = (priority,item)
      heapq.heappush(self.heap,pair)

  def pop(self):
      (priority,item) = heapq.heappop(self.heap)
      return item
  
  def isEmpty(self):
    return len(self.heap) == 0



class SpaceStation:

    def __init__(self, size = 5):
        self.size = size
        midpoint = round(float(size) / 2.0 + .01) - 1
        self.airlock = [midpoint, midpoint]
    
    def getMoves(self, game, i):
        y = game[i][0]
        x = game[i][1]
        successors = []
        # check north
        loop = True
        yt = y - 1
        if [yt, x] in game:
            loop = False
        while loop and yt > 0:
            if [yt - 1, x] in game:
                tmp = list(game)
                if [yt, x] == self.airlock and i != 0:
                    tmp.pop(i)
                else:
                    tmp[i] = [yt, x]
                successors.append([tmp, [i, 'North']])
                loop = False
            yt -= 1

        # check south
        loop = True
        yt = y + 1
        if [yt, x] in game:
            loop = False
        while loop and yt < self.size:
            if [yt + 1, x] in game:
                tmp = list(game)
                if [yt, x] == self.airlock and i != 0:
                    tmp.pop(i)
                else:
                    tmp[i] = [yt, x]
                successors.append([tmp, [i, 'South']])
                loop = False
            yt += 1


        # check east
        loop = True
        xt = x + 1
        if [y, xt] in game:
            loop = False
        while loop and xt < self.size:
            if [y, xt + 1] in game:
                tmp = list(game)
                if [y, xt] == self.airlock and i != 0:
                    tmp.pop(i)
                else:
                    tmp[i] = [y, xt]
                successors.append([tmp, [i, 'East']])
                loop = False
            xt += 1


        # check west
        loop = True
        xt = x - 1
        if [y, xt] in game:
            loop = False
        while loop and xt > 0:
            if [y, xt - 1] in game:
                tmp = list(game)
                if [y, xt] == self.airlock and i != 0:
                    tmp.pop(i)
                else:
                    tmp[i] = [y, xt]
                successors.append([tmp,  [i, 'West']])
                loop = False
            xt -= 1

        return successors
            

    def getSuccessors(self, game):
        """ given a game in the form of
        [[1,1], [0, 1], [0, 2], [3, 3], [4,1]]
        will return a list of successors"""
        
        successors = []
        for i in range(len(game)):
            successors += self.getMoves(game, i)
        return successors

    def getSolution(self, game):
        return solution        


# your code should pass these tests

# The following is one solution for game 1. There is at least one more.
#solution1 = [[0, 'South'], [0, 'East'], [0, 'North'], [1, 'South'], [3, 'West'], [0, 'South']]
game1 =  [[1,1], [0, 1], [0, 2], [3, 3], [4,1]]

#solution2 = [[2, 'West'], [5, 'North'], [5, 'East'], [5, 'South'], [5, 'West'], [4, 'North'], [4, 'East'], [0, 'North'], [0, 'West']]

game2 = [[3,3], [0,0], [0, 2], [1, 4], [2, 0], [4, 1]]


spaceStation = SpaceStation()
print spaceStation.getSolution(game1)
